@@ -1,6 +1,6 @@
 /***********************************************************************/
 /*                                                                     */
-/*                           Objective Caml                            */
+/*                                OCaml                                */
 /*                                                                     */
 /*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         */
 /*                                                                     */
@@ -11,8 +11,9 @@
 /*                                                                     */
 /***********************************************************************/
 
-/* $Id: fchown.c,v 1.8 2001/12/07 13:40:28 xleroy Exp $ */
+/* $Id: fchown.c 11156 2011-07-27 14:17:02Z doligez $ */
 
+#include <fail.h>
 #include <mlvalues.h>
 #include "unixsupport.h"
 
@@ -29,5 +30,5 @@ CAMLprim value unix_fchown(value fd, value uid, value gid)
 
 CAMLprim value unix_fchown(value fd, value uid, value gid)
 { invalid_argument("fchown not implemented"); }
-  
+
 #endif
