@@ -1,6 +1,6 @@
 /***********************************************************************/
 /*                                                                     */
-/*                           Objective Caml                            */
+/*                                OCaml                                */
 /*                                                                     */
 /*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         */
 /*                                                                     */
@@ -11,8 +11,9 @@
 /*                                                                     */
 /***********************************************************************/
 
-/* $Id: getpeername.c,v 1.10 2004/04/09 13:25:21 xleroy Exp $ */
+/* $Id: getpeername.c 11156 2011-07-27 14:17:02Z doligez $ */
 
+#include <fail.h>
 #include <mlvalues.h>
 #include "unixsupport.h"
 
@@ -29,12 +30,12 @@ CAMLprim value unix_getpeername(value sock)
   addr_len = sizeof(addr);
   retcode = getpeername(Int_val(sock), &addr.s_gen, &addr_len);
   if (retcode == -1) uerror("getpeername", Nothing);
-  return alloc_sockaddr(&addr, addr_len);
+  return alloc_sockaddr(&addr, addr_len, -1);
 }
 
 #else
 
 CAMLprim value unix_getpeername(value sock)
 { invalid_argument("getpeername not implemented"); }
-  
+
 #endif
