@@ -1,6 +1,6 @@
 /***********************************************************************/
 /*                                                                     */
-/*                           Objective Caml                            */
+/*                                OCaml                                */
 /*                                                                     */
 /*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         */
 /*                                                                     */
@@ -11,12 +11,12 @@
 /*                                                                     */
 /***********************************************************************/
 
-/* $Id: getppid.c,v 1.8 2001/12/07 13:40:30 xleroy Exp $ */
+/* $Id: getppid.c 11156 2011-07-27 14:17:02Z doligez $ */
 
 #include <mlvalues.h>
 #include "unixsupport.h"
 
-CAMLprim value unix_getppid(void)
+CAMLprim value unix_getppid(value unit)
 {
   return Val_int(getppid());
 }
