@@ -1,6 +1,6 @@
 /***********************************************************************/
 /*                                                                     */
-/*                           Objective Caml                            */
+/*                                OCaml                                */
 /*                                                                     */
 /*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         */
 /*                                                                     */
@@ -11,8 +11,9 @@
 /*                                                                     */
 /***********************************************************************/
 
-/* $Id: rewinddir.c,v 1.11 2004/06/19 15:38:31 xleroy Exp $ */
+/* $Id: rewinddir.c 11156 2011-07-27 14:17:02Z doligez $ */
 
+#include <fail.h>
 #include <mlvalues.h>
 #include "unixsupport.h"
 #include <errno.h>
