@@ -1,6 +1,6 @@
 /***********************************************************************/
 /*                                                                     */
-/*                           Objective Caml                            */
+/*                                OCaml                                */
 /*                                                                     */
 /*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         */
 /*                                                                     */
@@ -11,14 +11,14 @@
 /*                                                                     */
 /***********************************************************************/
 
-/* $Id: time.c,v 1.9 2001/12/07 13:40:36 xleroy Exp $ */
+/* $Id: time.c 11156 2011-07-27 14:17:02Z doligez $ */
 
 #include <time.h>
 #include <mlvalues.h>
 #include <alloc.h>
 #include "unixsupport.h"
 
-CAMLprim value unix_time(void)
+CAMLprim value unix_time(value unit)
 {
   return copy_double((double) time((time_t *) NULL));
 }
