@@ -1,6 +1,6 @@
 /***********************************************************************/
 /*                                                                     */
-/*                           Objective Caml                            */
+/*                                OCaml                                */
 /*                                                                     */
 /*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         */
 /*                                                                     */
@@ -12,7 +12,7 @@
 
 /* Based on public-domain code from Berkeley Yacc */
 
-/* $Id: reader.c,v 1.28 2004/06/12 11:59:11 xleroy Exp $ */
+/* $Id: reader.c 11156 2011-07-27 14:17:02Z doligez $ */
 
 #include <string.h>
 #include "defs.h"
@@ -163,6 +163,7 @@ char *substring (char *str, int start, int len)
   for (i = 0; i < len; i++){
     buf[i] = str[start+i];
   }
+  buf[i] = '\0';      /* PR#4796 */
   return buf;
 }
 
@@ -710,7 +711,7 @@ get_literal(void)
     n = cinc;
     s = MALLOC(n);
     if (s == 0) no_space();
-    
+
     for (i = 0; i < n; ++i)
         s[i] = cache[i];
 
@@ -825,13 +826,17 @@ get_tag(void)
     register int i;
     register char *s;
     char *t_line = dup_line();
+    long bracket_depth;
 
     cinc = 0;
+    bracket_depth = 0;
     while (1) {
       c = *++cptr;
       if (c == EOF) unexpected_EOF();
       if (c == '\n') syntax_error(lineno, line, cptr);
-      if (c == '>' && cptr[-1] != '-') break;
+      if (c == '>' && 0 == bracket_depth && cptr[-1] != '-') break;
+      if (c == '[') ++ bracket_depth;
+      if (c == ']') -- bracket_depth;
       cachec(c);
     }
     ++cptr;
@@ -1266,9 +1271,9 @@ void copy_action(void)
             nrules-2, input_file_name, lineno);
             */
     if (sflag)
-      fprintf(f, "yyact.(%d) <- (fun parser_env ->\n", nrules-2);
+      fprintf(f, "yyact.(%d) <- (fun __caml_parser_env ->\n", nrules-2);
     else
-      fprintf(f, "; (fun parser_env ->\n");
+      fprintf(f, "; (fun __caml_parser_env ->\n");
 
     n = 0;
     for (i = nitems - 1; pitem[i]; --i) ++n;
@@ -1278,11 +1283,13 @@ void copy_action(void)
       if (item->class == TERM && !item->tag) continue;
       fprintf(f, "    let _%d = ", i);
       if (item->tag)
-        fprintf(f, "(peek_val parser_env %d : %s) in\n", n - i, item->tag);
+        fprintf(f, "(Parsing.peek_val __caml_parser_env %d : %s) in\n", n - i,
+                item->tag);
       else if (sflag)
-        fprintf(f, "peek_val parser_env %d in\n", n - i);
+        fprintf(f, "Parsing.peek_val __caml_parser_env %d in\n", n - i);
       else
-        fprintf(f, "(peek_val parser_env %d : '%s) in\n", n - i, item->name);
+        fprintf(f, "(Parsing.peek_val __caml_parser_env %d : '%s) in\n", n - i,
+                item->name);
     }
     fprintf(f, "    Obj.repr(\n");
     fprintf(f, line_format, lineno, input_file_name);
@@ -1300,7 +1307,7 @@ loop:
         {
             ++cptr;
             i = get_number();
-            
+
             if (i <= 0 || i > n)
               unknown_rhs(i);
             item = pitem[nitems + i - n - 1];
@@ -1387,7 +1394,7 @@ loop:
           fwrite(cptr, 1, 2, f);
           cptr += 2;
         } else
-        if (cptr[0] == '\\' 
+        if (cptr[0] == '\\'
             && isdigit((unsigned char) cptr[1])
             && isdigit((unsigned char) cptr[2])
             && isdigit((unsigned char) cptr[3])
@@ -1724,7 +1731,12 @@ static int is_polymorphic(char * s)
 {
   while (*s != 0) {
     char c = *s++;
-    if (c == '\'') return 1;
+    if (c == '\'' || c == '#') return 1;
+    if (c == '[') {
+      c = *s;
+      while (c == ' ' || c == '\t' || c == '\r' || c == '\n') c = *++s;
+      if (c == '<' || c == '>') return 1;
+    }
     if (In_bitmap(caml_ident_start, c)) {
       while (In_bitmap(caml_ident_body, *s)) s++;
     }
@@ -1758,7 +1770,7 @@ void make_goal(void)
       if (is_polymorphic(bp->tag))
         polymorphic_entry_point(bp->name);
       fprintf(entry_file,
-              "let %s (lexfun : Lexing.lexbuf -> token) (lexbuf : Lexing.lexbuf) =\n   (yyparse yytables %d lexfun lexbuf : %s)\n",
+              "let %s (lexfun : Lexing.lexbuf -> token) (lexbuf : Lexing.lexbuf) =\n   (Parsing.yyparse yytables %d lexfun lexbuf : %s)\n",
               bp->name, bp->entry, bp->tag);
       fprintf(interface_file,
               "val %s :\n  (Lexing.lexbuf  -> token) -> Lexing.lexbuf -> %s\n",
@@ -1768,11 +1780,13 @@ void make_goal(void)
               "(* Entry %s *)\n", bp->name);
       if (sflag)
         fprintf(action_file,
-                "yyact.(%d) <- (fun parser_env -> raise (YYexit (peek_val parser_env 0)))\n",
+                "yyact.(%d) <- (fun __caml_parser_env -> raise "
+                "(Parsing.YYexit (Parsing.peek_val __caml_parser_env 0)))\n",
                 ntotalrules);
       else
         fprintf(action_file,
-              "; (fun parser_env -> raise (YYexit (peek_val parser_env 0)))\n");
+                "; (fun __caml_parser_env -> raise "
+                "(Parsing.YYexit (Parsing.peek_val __caml_parser_env 0)))\n");
       ntotalrules++;
       last_was_action = 1;
       end_rule();
